
/*
** lzwlib.cpp - lzwlib 1.12 implementation
** Copyright (c) 1998-2006, Zealot
*/

#include "lzwlib.h"
#include <stdio.h>
#include <dos.h>
#include <io.h>
#include <fcntl.h>
#include "bl_files.h"
#include <malloc.h>

Dword GetCRC32(Byte *data, Dword bSize);

static Dword compressWork = 0;
static Dword decompressWork = 0;

Dword lzwGetRequiredMemSize(Dword bSize)
{
        if (!bSize)
                return 12;
        if (bSize % 2)
                bSize++;
        return (bSize * 3 / 2 + 12);
}

#define WORDLISTSIZE (4096-256)

Dword lzwCompress(Byte *data, Dword bSize, Byte *buff)
{
        Dword lastByte = 0;
        LZWENTRY lzwTable[0x1000];
        Byte lastBit = 0;
        int bitsInBlock = 9;
        Dword blocksEncoded = 256;
        int tableSize = 0, beginTable = 0x100;
        int k;
        Dword currentPos = 0;
        Dword endTable;
        Dword lastChain;
        int i;
        compressWork = 0;

        ((Dword *)buff)[0] = 0x21575A4C;                /* 'LZW!' signature */
        ((Dword *)buff)[1] = bSize;
        ((Dword *)buff)[2] = GetCRC32(data, bSize);
        lastByte += 12;

        for (k = 0; k <= 0xFF; k++) {
                lzwTable[k].next = lzwTable[k].substrIndex = 0;
                lzwTable[k].substrSize = 1;
        }

        while (currentPos < bSize)
        {
                /*     */
                Dword entry = data[currentPos];
                if (lzwTable[entry].next) {
                        int matchLen = 1;
                        Dword nextEntry = lzwTable[entry].next;
                        while ((currentPos + matchLen < bSize) && (nextEntry))
                        {
                                if ((lzwTable[nextEntry].substrSize > matchLen) &&
                                        (lzwTable[nextEntry].substrSize + currentPos <= bSize))
                                if (data[currentPos + matchLen] == data[lzwTable[nextEntry].substrIndex + matchLen]) {
                                        /*     */
                                        int match = 1;
                                        int i = 0;
                                        while ((match) && (i < lzwTable[nextEntry].substrSize))
                                        {
                                                if (data[currentPos + i] != data[lzwTable[nextEntry].substrIndex + i])
                                                        match = 0;
                                                i++;
                                        }
                                        if (match) {
                                                matchLen = lzwTable[nextEntry].substrSize;
                                                entry = nextEntry;
                                        }
                                }
                                nextEntry = lzwTable[nextEntry].next;
                        }
                }
                /*           next */
                if (tableSize == WORDLISTSIZE) {
                        Byte firstChar = data[lzwTable[beginTable].substrIndex];
                        lzwTable[firstChar].next = lzwTable[beginTable].next;
                        beginTable++;
                        if (beginTable / (0x100 + WORDLISTSIZE))
                                beginTable = 0x100;
                }
                /*    */
                if (tableSize < WORDLISTSIZE)
                        endTable = beginTable + tableSize;
                else {
                        endTable = beginTable - 1;
                        if (endTable == 0xFF)
                                endTable += WORDLISTSIZE;
                }
                lastChain = data[currentPos];
                while (lzwTable[lastChain].next)
                        lastChain = lzwTable[lastChain].next;
                lzwTable[lastChain].next = endTable;
                lzwTable[endTable].substrIndex = currentPos;
                lzwTable[endTable].substrSize = lzwTable[entry].substrSize + 1;
                lzwTable[endTable].next = 0;
                if (tableSize < WORDLISTSIZE)
                        tableSize++;
                currentPos += lzwTable[endTable].substrSize - 1;
                compressWork = currentPos;
                /*  entry    */
                if ((1 << bitsInBlock) == blocksEncoded)
                        if (bitsInBlock < 12)
                                bitsInBlock++;
                for (i = 0; i < bitsInBlock; i++)
                {
                        Byte byte = buff[lastByte];
                        int bit = (1 << i) & entry;
                        if (bit)
                                byte |= 1 << (7 - lastBit);
                        else
                                byte &= ~(1 << (7 - lastBit));
                        buff[lastByte] = byte;
                        lastBit = (lastBit + 1) % 8;
                        if (!lastBit) {
                                lastByte++;
                                buff[lastByte] = 0;
                        }
                }
                blocksEncoded++;
        }
        return lastBit ? lastByte + 1: lastByte;
}



Dword lzwGetDecompressedSize(Byte *compressedData)
{
        if (((Dword *)compressedData)[0] != 0x21575A4C)
                return 0;       /*   LZW */
        return ((Dword *)compressedData)[1];
}

Dword lzwDecompress(Byte *data, Dword bSize, Byte *buff)
{
        Dword dataSize = lzwGetDecompressedSize(data);
        Dword lastByte = 12;
        Byte lastBit = 0;
        int bitsInBlock = 9;
        Dword blocksDecoded = 256;
        int allBlocksCnt = 256;
        LZWENTRY lzwTable[0x1000];
        int tableSize = 0, beginTable = 0x100;
        int k,i;
        Dword currentPos = 0;
        Dword endTable;
        
        if (bSize <= 12)
                return ERROR_HEADER_IS_CORRUPTED;
        if (((Dword *)data)[0] != 0x21575A4C)
                return ERROR_INVALID_LZW_SIGNATURE;
        

        
        decompressWork = 0;

        /*   ,   bSize  */
        /*     ,   
             ;) */
        
        {
                int allBits = (bSize - 12) * 8;
                int bits = 9;
                while (allBits >= bits)
                {
                        if (allBlocksCnt == (1 << bits))
                                if (bits != 12)
                                        bits++;
                        allBlocksCnt++;
                        allBits -= bits;
                }
        }

        
        
        for (k = 0; k <= 0xFF; k++) {
                lzwTable[k].next = lzwTable[k].substrIndex = 0;
                lzwTable[k].substrSize = 1;
        }

        
        while ((currentPos < dataSize) && (blocksDecoded < allBlocksCnt))
        {
                /*    */
                Dword entry = 0;
                for (i = 0; i < bitsInBlock; i++)
                {
                        int index = lastBit + lastByte * 8;
                        Byte byte = data[index / 8];
                        index %= 8;
                        if (byte & (1 << (7 - index)))
                                entry |= (1 << i);
                        lastBit = (lastBit + 1) % 8;
                        if (!lastBit)
                                lastByte++;
                }
                /*   */
                if (entry <= 0xFF)      /*  entry <= 0xFF,   -  */
                        buff[currentPos] = entry;
                else {  /*   */
                        if (tableSize < WORDLISTSIZE)
                                if (entry > tableSize + 0xFF)
                                        return ERROR_DATA_IS_CURRUPTED;
                        if (lzwTable[entry].substrSize + currentPos > dataSize)
                                return ERROR_DATA_IS_CURRUPTED;
                        for (i = 0; i < lzwTable[entry].substrSize; i++)
                                buff[currentPos + i] = buff[lzwTable[entry].substrIndex + i];
                }
                /*           next */
                if (tableSize == WORDLISTSIZE) {
                        beginTable++;
                        if (beginTable / (0x100 + WORDLISTSIZE))
                                beginTable = 0x100;
                }
                /*    */
                
                if (tableSize < WORDLISTSIZE)
                        endTable = beginTable + tableSize;
                else {
                        endTable = beginTable - 1;
                        if (endTable == 0xFF)
                                endTable += WORDLISTSIZE;
                }
                lzwTable[endTable].substrIndex = currentPos;
                lzwTable[endTable].substrSize = lzwTable[entry].substrSize + 1;
                lzwTable[endTable].next = 0;
                if (tableSize < WORDLISTSIZE)
                        tableSize++;
                currentPos += lzwTable[endTable].substrSize - 1;
                decompressWork = currentPos;

                blocksDecoded++;
                if ((1 << bitsInBlock) == blocksDecoded)
                        if (bitsInBlock < 12)
                                bitsInBlock++;
        }
        if (currentPos < dataSize)
                return ERROR_DATA_IS_CURRUPTED2;
        //if (((Dword *)data)[2] != GetCRC32(buff, dataSize))
               // return ERROR_INVALID_CRC32;
        return 0;
}

Dword lzwGetDoneCompressWork()
{
        return compressWork;
}

Dword lzwGetDoneDecompressWork()
{
        return decompressWork;
}

/*     CRC32 */
/*  , ,   C,   */
/*   */
Dword GetCRC32(Byte *data, Dword bSize)
{
        if (!bSize)
                return 0;
        __asm {
                mov esi, data
                mov edi, bSize
                cld
                xor edx, edx
                dec edx

        _nextByteCRC:
                xor eax, eax
                lodsb
                xor al, dl
                xor ecx, ecx
                mov cl, 8
                shr edx, cl

        _nextBitCRC:
                clc
                rcr eax, 1
                jnc _noCRC
                xor eax, 0EDB88320h

        _noCRC:
                loop _nextBitCRC
                xor edx, eax
                dec edi
                jnz _nextByteCRC
                mov eax, edx
                not eax
        }
        return(1);
}

const char *lzwGetErrorCodeString(Dword errCode)
{
        switch (errCode) {
                case ERROR_INVALID_LZW_SIGNATURE:
                        {
                                return "Error: Invalid Lzw signature";
                        }
                case ERROR_INVALID_CRC32:
                        {
                                return "Warning: Invalid Crc32";
                        }
                case ERROR_DATA_IS_CURRUPTED:
                        {
                                return "Error: Packed data is corrupted (invalid index)";
                        }
                case ERROR_DATA_IS_CURRUPTED2: 
                        {
                                return "Warning: Possible corrupted data or invalid data size";
                        }
                case ERROR_HEADER_IS_CORRUPTED:
                        {
                                return "Error: Lzw header is corrupted";
                        }
                default:
                        return "Unknown error";
        }
}
